(******************************************************************************)
(** * Specialized list filters    *)
(******************************************************************************)

Require Import Omega.
Require Import Hahn.
Require Import List.
Require Import Basic.

Set Implicit Arguments.
Remove Hints plus_n_O.

Section Filter.

Variable A : Type.

(******************************************************************************)
(** Definitions   *)
(******************************************************************************)

Fixpoint filter_ind (l : list A) (P : list bool) : list A :=
  match l,P with
  | nil,_ => nil
  | _,nil => nil
  | x :: l, false :: P => x :: filter_ind l P
  | x :: l, true :: P => filter_ind l P
  end.

Definition filter_ind_map (P : list bool) (n : nat) :=
  length (filter (Bool.eqb false) (firstn n P)).

Fixpoint filter_ind_map_inv (P : list bool) (n : nat) :=
  match P,n with
  | nil,_ => 1
  | false :: P,0 => 0
  | false :: P,S n => S (filter_ind_map_inv P n)
  | true :: P,n => S (filter_ind_map_inv P n)
  end.

Fixpoint min_of_P (P : list bool) :=
  match P with
    | nil => 0
    | b :: p' => if b then 0
                 else S (min_of_P p')
  end.

Definition filter_ind_pre (l : list A) (P : list bool) : list A :=
  skipn (min_of_P P) (filter_ind l P).

Definition filter_ind_pre_map (P : list bool) n :=
  filter_ind_map P n - min_of_P P.

Definition filter_ind_pre_map_inv (P : list bool) (n : nat) :=
  filter_ind_map_inv P (n + min_of_P P).

(******************************************************************************)
(** Examples - uncomment to play around   *)
(******************************************************************************)

(*
End Filter.

Import ListNotations.

Compute (filter_ind [11 ; 21 ; 31 ; 41 ; 51] [false ; true ; false ; true ; false]).
Compute (filter_ind_map [false ; true ; false ; true ; false] 4).
Compute (filter_ind_map_inv [false ; true ; false ; true ; false] 2).

Compute (filter_ind_pre_map [false ; true ; false ; true ; false] 2).
Compute (filter_ind_pre_map_inv [false ; true ; false ; true ; false] 0).

Compute (filter_ind_pre_map [true ; true ; false ; true ; false] 2).
Compute (filter_ind_pre_map_inv [true ; true ; false ; true ; false] 0).


Compute (filter_ind [11 ; 21 ; 31 ; 41 ; 51] [false ; true ; false ; true ; false]).
Compute (filter_ind_map [false ; true ; false ; true ; false] 4).

Compute (filter_ind_pre [11 ; 21 ; 31 ; 41 ; 51] [false ; true ; false ; true ; false]).
Compute (filter_ind_pre_map [false ; true ; false ; true ; false] 4).
Compute (filter_ind_pre [11 ; 21 ; 31 ; 41 ; 51] [false ; false ; false ; false ; false]).
Compute (min_of_P [false ; false ; false ; false ; false]).
Compute (filter_ind_map [false ; false ; false ; false ; false] 4).
Compute (filter_ind [11 ; 21 ; 31 ; 41 ; 51] [false ; true ; false ; true ; false]).
Compute (filter_ind [11 ; 21 ; 31 ; 41 ; 51] [false ; true ; false ; true ; false]).
Compute (filter_ind_map [false ; true ; false ; true ; false] 4).

Compute (filter_ind_pre [11 ; 21 ; 31 ; 41 ; 51] [false ; false ; false ; false ; false]).
*)

(******************************************************************************)
(** Properties of min_of_P   *)
(******************************************************************************)

Lemma false_before_min (P : list bool) n
      (LEQ: n < min_of_P P) : 
      nth n P false = false.
Proof.
generalize dependent n.
induction P; ins; desf; try omega.
apply IHP; omega.
Qed.

Lemma true_in_min (P : list bool) 
  (IN:  min_of_P P < length P) :
  nth (min_of_P P) P false = true.
Proof.
induction P; ins; desf.
omega.
apply IHP; omega.
Qed.

Lemma P_non_empty_min (P : list bool) :
      (exists m, nth_error P m = Some true) <-> 
      min_of_P P < length P.
Proof.
split; ins; desc.
- generalize dependent m.
induction P; ins; desf.
destruct m; desf.
omega.
destruct m; desf; ins.
apply IHP in H.
omega.
- exists (min_of_P P).
erewrite nth_error_nth'; try done.
rewrite true_in_min; try done.
Qed.

Lemma min_length P : min_of_P P <= length P.
Proof.
induction P; desf.
simpl; desf; omega.
Qed.

Lemma min_before_true (P : list bool) 
  n (MIN: nth n P false = true) :
  min_of_P P <= n.
Proof.
generalize dependent n.
induction P; ins; desf.
omega.
apply le_n_S.
auto.
Qed.

Lemma min_of_P_alt 
  (l : list bool) n 
  (IN:  min_of_P l < length l) :
  (nth n l false = true /\
  forall m, m < n -> nth m l false = false) <->
  min_of_P l = n.
Proof.
split; intro.
generalize dependent l.
- induction n; ins.
+ simpl in H; desf.
destruct l; desf.
destruct b; desf.
+ 
destruct l; desf.
destruct b; desf.
by specialize (H0 0 (Nat.lt_0_succ n)); simpl in H0; desf.
simpl.
apply f_equal.
eapply IHn; splits; eauto.
by simpl in IN; omega.
intros m L.
specialize (H0 (S m) (lt_n_S _ _ L)).
by simpl in H0.
- split; subst.
by eapply true_in_min.
by ins; apply false_before_min.
Qed.

(******************************************************************************)
(** Properties of filter_ind and filter_ind_pre  *)
(******************************************************************************)

Lemma filter_ind_P_empty_min l (P : list bool)
  (LEN: length l = length P)
  (EMP: ~ (exists m, nth_error P m = Some true)) :
  filter_ind l P = l.
Proof.
generalize dependent P.
induction l; ins.
destruct P eqn:B; desf.
exfalso; eapply EMP.
exists 0; ins.
rewrite IHl; ins.
omega.
intro; desc; eapply EMP.
exists (S m); ins.
Qed.

Lemma filter_ind_pre_P_empty_min l (P : list bool)
  (LEN: length l = length P)
  (EMP: ~ (exists m, nth_error P m = Some true)) :
  filter_ind_pre l P = nil.
Proof.
generalize dependent P.
induction l; ins.
by destruct P; desf.
destruct P eqn:B; desf.
unfold filter_ind_pre; desf.
destruct b.
exfalso; eapply EMP.
exists 0; ins.
simpl.
apply IHl; ins.
omega.
intro; desc; eapply EMP.
exists (S m); ins.
Qed.

(******************************************************************************)
(** Properties of filter_ind_map   *)
(******************************************************************************)

Lemma filter_ind_map_prop (l : list A) (P : list bool) 
      (LEN: length l = length P) n (NIN: nth_error P n = Some false) :
      nth_error l n = nth_error (filter_ind l P) (filter_ind_map P n).
Proof.
generalize dependent l.
generalize dependent P.
induction n.
- 
unfold filter_ind.
ins; desf.
- 
unfold filter_ind_map.
destruct l, P; desf.
ins.
desf; ins; eauto.
Qed.

Lemma filter_ind_map_prop' (l : list A) (P : list bool) 
      (LEN: length l = length P) n (NIN: n < length l)
      (NINP: nth n P false = false) def :
      nth n l def = nth (filter_ind_map P n) (filter_ind l P) def.
Proof.
assert (nth_error P n = Some false).
  { rewrite LEN in NIN.
    apply nth_error_nth' with (d:=false) in NIN.
    congruence.
  }
destruct (nth_error l n) eqn:E.
- assert (E':=E).
erewrite filter_ind_map_prop in E; try edone.
apply nth_error_nth with (d:=def) in E.
apply nth_error_nth with (d:=def) in E'.
congruence.
- assert (E':=E).
erewrite filter_ind_map_prop in E; try edone.
apply nth_error_nth_d with (d:=def) in E.
apply nth_error_nth_d with (d:=def) in E'.
congruence.
Qed.


Lemma filter_ind_map_monotone_succ (P : list bool) 
  n : filter_ind_map P n <= filter_ind_map P (S n).
Proof.
unfold filter_ind_map.
generalize dependent P.
induction n.
- destruct P; ins; omega.
- intros P.
destruct P.
by ins.
remember (S n) as m.
ins; destruct b; ins.
by subst m; ins.
subst m; ins.
by apply le_n_S.
Qed. 

Lemma filter_ind_map_monotone (P : list bool) 
  n1 n2 (LEQ: n1 <= n2) : filter_ind_map P n1 <= filter_ind_map P n2.
Proof.
induction LEQ.
auto.
rewrite IHLEQ; apply filter_ind_map_monotone_succ.
Qed.

Lemma filter_ind_map_monotone_strict (P : list bool) 
  n1 n2 (LT: n1 < n2) (LEN: n2 < length P)
  (NINP: nth n1 P false = false) :
  filter_ind_map P n1 < filter_ind_map P n2.
Proof.
unfold filter_ind_map.
generalize dependent n1.
generalize dependent n2.
induction P.
by ins; desf; omega.
ins; destruct n1, n2; try omega.
- simpl; desf; ins; omega.
- simpl; desf; simpl.
apply lt_n_S.
eapply IHP; try done; omega.
eapply IHP; try done; omega.
Qed.

Lemma filter_ind_map_length (l : list A) (P : list bool) 
 n (IN: n < length l) (LEN: length l = length P) 
 (NINP: nth n P false = false):
filter_ind_map P n < length (filter_ind l P).
Proof.
apply nth_error_Some.
rewrite <- filter_ind_map_prop with (P:=P); try done.
rewrite nth_error_None.
omega.
erewrite nth_error_nth_len.
by rewrite NINP.
by rewrite <- LEN.
Qed.

(******************************************************************************)
(** Properties of filter_ind_pre_map   *)
(******************************************************************************)

Lemma filter_ind_pre_length l (P : list bool) 
      (LEN: length l = length P) :
 length (filter_ind_pre l P) + (min_of_P P) = length (filter_ind l P).
Proof.
generalize dependent l.
induction P.
by destruct l; ins; omega.
ins; desf; unfold filter_ind_pre.
- destruct l; ins; omega.
- destruct l; ins.
rewrite Nat.add_succ_r.
apply eq_S, IHP; omega.
Qed.

Lemma filter_ind_pre_map_min (P : list bool) 
 n (LEQ : min_of_P P <= n) :
min_of_P P <= filter_ind_map P n.
Proof.
unfold filter_ind_map.
generalize dependent n.
induction P.
- ins; omega.
- ins; desf.
omega.
destruct n; desf.
ins.
apply le_n_S.
apply IHP.
omega.
Qed.

Lemma filter_ind_pre_map_prop (l : list A) (P : list bool) 
      (LEN: length l = length P) 
      n (NIN: nth_error P n = Some false) (LEQ: min_of_P P <= n) :
      nth_error l n = nth_error (filter_ind_pre l P) (filter_ind_pre_map P n).
Proof.
unfold filter_ind_pre.
rewrite skipn_nth_error.
unfold filter_ind_pre_map.
erewrite filter_ind_map_prop; try edone.
rewrite le_plus_minus_r; [done|].
apply filter_ind_pre_map_min; try done.
Qed.

Lemma filter_ind_pre_map_prop' (l : list A) (P : list bool) 
      (LEN: length l = length P) 
      n (NIN: n < length l)  (LEQ: min_of_P P <= n) 
      (NINP: nth n P false = false) def :
      nth n l def = nth (filter_ind_pre_map P n) (filter_ind_pre l P) def.
Proof.
assert (nth_error P n = Some false).
  { rewrite LEN in NIN.
    apply nth_error_nth' with (d:=false) in NIN.
    congruence.
  }
destruct (nth_error l n) eqn:E.
- assert (E':=E).
erewrite filter_ind_pre_map_prop in E; try edone.
apply nth_error_nth with (d:=def) in E.
apply nth_error_nth with (d:=def) in E'.
congruence.
- assert (E':=E).
erewrite filter_ind_pre_map_prop in E; try edone.
apply nth_error_nth_d with (d:=def) in E.
apply nth_error_nth_d with (d:=def) in E'.
congruence.
Qed.

Lemma filter_ind_pre_map_length (l : list A) (P : list bool) 
 n (IN: n < length l) (LEN: length l = length P) 
 (LEQ: min_of_P P <= n) (NINP: nth n P false = false):
filter_ind_pre_map P n < length (filter_ind_pre l P).
Proof.
apply nth_error_Some.
rewrite <- filter_ind_pre_map_prop with (P:=P); try done.
rewrite nth_error_None.
- omega.
- erewrite nth_error_nth_len.
by rewrite NINP.
by rewrite <- LEN.
Qed.

Lemma filter_ind_pre_map_monotone_strict (P : list bool) 
  n1 n2 (LT: n1 < n2) (LEN: n2 < length P)
  (MIN: min_of_P P <= n1)
  (NINP1: nth n1 P false = false) :
  filter_ind_pre_map P n1 < filter_ind_pre_map P n2.
Proof.
unfold filter_ind_pre_map.
exploit (@filter_ind_map_monotone_strict P n1 n2); try edone.
intro.
exploit filter_ind_pre_map_min; try edone.
ins; omega.
Qed.

(******************************************************************************)
(** filter_ind_map_inv is inverse of filter_ind_map   *)
(******************************************************************************)

Lemma filter_ind_map_inv_l P n 
(NIN: n < length P) (NINP: nth n P false = false) :
filter_ind_map_inv P (filter_ind_map P n) = n.
Proof.
generalize dependent n.
induction P.
ins; omega.
ins; desf.
- unfold filter_ind_map in *; ins.
rewrite IHP; try omega.
done.
- unfold filter_ind_map in *; ins.
apply eq_add_S in Heq.
rewrite <- Heq.
rewrite IHP; try omega.
done.
Qed.

Lemma filter_ind_map_inv_r l P n 
(LEN: length l = length P) (NIN: n < length (filter_ind l P)) :
filter_ind_map P (filter_ind_map_inv P n) = n.
Proof.
generalize dependent l.
generalize dependent n.
induction P.
- destruct l; ins; desf; omega.
- ins; desf.
+ unfold filter_ind_map in *; ins.
destruct l; ins.
apply eq_add_S in LEN.
erewrite IHP; try edone.
+ unfold filter_ind_map in *; ins.
destruct l; ins.
apply eq_add_S in LEN.
erewrite IHP; try edone.
omega.
Qed.

(******************************************************************************)
(** Properties of filter_ind_map_inv   *)
(******************************************************************************)

Lemma filter_ind_map_inv_length (l : list A) (P : list bool) 
 n (LEN: length l = length P)  (IN: n < length (filter_ind l P)) :
filter_ind_map_inv P n < length l.
Proof.
generalize dependent l.
generalize dependent n.
induction P.
- destruct l; ins; desf; omega.
- ins; desf; try omega.
+ 
destruct l; ins.
apply lt_n_S; eapply IHP; eauto.
+ destruct l; ins.
apply lt_n_S; eapply IHP; eauto.
omega.
Qed.

Lemma filter_ind_map_inv_nin_P (l : list A) (P : list bool) 
 n (LEN: length l = length P)  (IN: n < length (filter_ind l P)) :
 nth (filter_ind_map_inv P n) P false = false.
Proof.
generalize dependent l.
generalize dependent n.
induction P.
- destruct l; ins; desf; omega.
- ins; desf.
+ 
destruct l; ins.
eapply IHP; eauto.
+ destruct l; ins.
eapply IHP; eauto.
omega.
Qed.

Lemma filter_ind_map_inv_prop' (l : list A) (P : list bool) 
      (LEN: length l = length P) n 
      (NIN: n < length (filter_ind l P)) def :
       nth (filter_ind_map_inv P n) l def =
       nth n (filter_ind l P) def.
Proof.
erewrite filter_ind_map_prop' with (l:=l); try edone.
erewrite filter_ind_map_inv_r with (l:=l); eauto.
by eapply filter_ind_map_inv_length.
eby eapply filter_ind_map_inv_nin_P.
Qed.

Lemma filter_ind_map_inv_monotone_strict l (P : list bool) 
  (LEN: length l = length P)
  n1 n2 (LT: n1 < n2) (LEN': n2 < length (filter_ind l P)):
  filter_ind_map_inv P n1 < filter_ind_map_inv P n2.
Proof.
assert (LEN'': n1 < length (filter_ind l P)).
omega.
destruct (classic (filter_ind_map_inv P n1 < filter_ind_map_inv P n2)).
done.
apply not_lt in H.
assert (filter_ind_map_inv P n1 = filter_ind_map_inv P n2 \/
filter_ind_map_inv P n2 < filter_ind_map_inv P n1).
omega.
desf.
- eapply f_equal with (f:=filter_ind_map P) in H0.
erewrite !filter_ind_map_inv_r in H0; try edone.
omega.
- eapply filter_ind_map_monotone_strict in H0; try edone.
erewrite !filter_ind_map_inv_r in H0; try edone.
omega.
rewrite <- LEN.
eapply filter_ind_map_inv_length; auto.
eapply filter_ind_map_inv_nin_P; try edone.
Qed.


(******************************************************************************)
(** filter_ind_pre_map_inv is inverse of filter_ind_pre_map   *)
(******************************************************************************)

Lemma filter_ind_pre_map_inv_l P n 
(NIN: n < length P) (LEQ: min_of_P P <= n)
(NINP: nth n P false = false) :
filter_ind_pre_map_inv P (filter_ind_pre_map P n) = n.
Proof.
unfold filter_ind_pre_map_inv, filter_ind_pre_map.
rewrite Nat.sub_add.
by apply filter_ind_map_inv_l.
apply filter_ind_pre_map_min.
done.
Qed.

Lemma filter_ind_pre_map_inv_r l P n 
(LEN: length l = length P) 
(NIN: n < length (filter_ind_pre l P)) :
filter_ind_pre_map P (filter_ind_pre_map_inv P n) = n.
Proof.
unfold filter_ind_pre_map_inv, filter_ind_pre_map.
rewrite filter_ind_map_inv_r with (l:=l); try edone.
omega.
exploit (filter_ind_pre_length); try edone.
ins; omega.
Qed.

(******************************************************************************)
(** Properties of filter_ind_pre_map_inv   *)
(******************************************************************************)

Lemma filter_ind_pre_map_inv_length (l : list A) (P : list bool) 
 n (LEN: length l = length P) 
(NIN: n < length (filter_ind_pre l P)) :
filter_ind_pre_map_inv P n < length l.
Proof.
unfold filter_ind_pre_map_inv.
apply filter_ind_map_inv_length; try done.
exploit (filter_ind_pre_length); try edone.
ins; omega.
Qed.


Lemma filter_ind_pre_map_inv_nin_P (l : list A) (P : list bool) 
 n (LEN: length l = length P)
 (NIN: n < length (filter_ind_pre l P)) :
 nth (filter_ind_pre_map_inv P n) P false = false.
Proof.
unfold filter_ind_pre_map_inv.
eapply filter_ind_map_inv_nin_P; try edone.
exploit (filter_ind_pre_length); try edone.
ins; omega.
Qed.

Lemma filter_ind_pre_map_inv_prop' (l : list A) (P : list bool) 
 n (LEN: length l = length P)
 (NIN: n < length (filter_ind_pre l P)) def :
      nth (filter_ind_pre_map_inv P n) l def =
      nth n (filter_ind_pre l P) def.
Proof.
unfold filter_ind_pre_map_inv.
rewrite filter_ind_map_inv_prop'; try edone.
2: exploit (filter_ind_pre_length); try edone; ins; omega.
unfold filter_ind_pre.
rewrite skipn_nth.
by rewrite Nat.add_comm.
Qed.

Lemma filter_ind_pre_map_inv_monotone_strict l (P : list bool) 
  (LEN: length l = length P)
  n1 n2 (LT: n1 < n2) 
  (NIN: n2 < length (filter_ind_pre l P)) :
  filter_ind_pre_map_inv P n1 < filter_ind_pre_map_inv P n2.
Proof.
unfold filter_ind_pre_map_inv.
eapply filter_ind_map_inv_monotone_strict; try edone.
omega.
exploit (filter_ind_pre_length); try edone.
ins; omega.
Qed.

Lemma min_lt_filter_ind_pre_map_inv P n:
  min_of_P P < filter_ind_pre_map_inv P n.
Proof.
unfold filter_ind_pre_map_inv.
induction P.
ins; omega.
simpl; desf.
omega.
omega.
apply lt_n_S.
assert (n + (min_of_P P) = n0).
omega.
rewrite <- H.
apply IHP.
Qed.

End Filter. 
